from ._function import FunctionTool, function
from ._browser import BrowserTool
from ._agent import AgentTool, BrowserAgentTool

from ._agent import agent_wrapper as wrap

__all__ = [
    'wrap',
    'function',
    'FunctionTool',
    'BrowserTool',
    'AgentTool',
    'BrowserAgentTool'
]



from .core.tool import FunctionTool, BrowserTool, function, AgentTool, BrowserAgentTool
from .core.hitl import HITL
from .llm import OpenAI, LLM
from npiai.core.base import Context

__all__ = [
    'BrowserTool',
    'FunctionTool',
    'AgentTool',
    'BrowserAgentTool',
    'function',
    'HITL',
    'OpenAI',
    'LLM',
    'Context'
]



"""This module contains the classes for the context and context message"""
import datetime
import uuid
import asyncio
from typing import List, Union

from fastapi import Request
from litellm.types.completion import ChatCompletionMessageParam


class Task:
    """the message wrapper of a message in the context"""

    def __init__(self, goal: str) -> None:
        self.task_id = str(uuid.uuid4())
        self.goal = goal
        self.born_at = datetime.datetime.now()
        self.dialogues: List[Union[ChatCompletionMessageParam]] = []
        self.response: str
        self._session: 'Context' | None = None

    async def step(self, msgs: List[Union[ChatCompletionMessageParam]]) -> None:
        """add a message to the context"""
        self.dialogues.extend(msgs)

        # for msg in msgs:
        #     if msg.get('content'):
        #         await self._session.send(callback.Callable(msg.get('content')))

    def set_session(self, session: 'Context') -> None:
        self._session = session

    def conversations(self) -> List[ChatCompletionMessageParam]:
        """return the raw message"""
        return self.dialogues.copy()


class Context:
    __last_screenshot: str | None = None

    def __init__(self, req: Request | None = None) -> None:
        # self._parent_session = parent
        # self._task = task
        self.id = str(uuid.uuid4())
        self.q = asyncio.Queue()
        # self.cb_dict: dict[str, callback.Callable] = {}
        self.__is_finished = False
        self.__result: str = ''
        self.__is_failed = False
        self.__failed_msg: str = ''
        self.__active_tool = None
        self.__request = req

    def authorization(self) -> str:
        return self.__request.headers.get('Authorization')

    def entry(self):
        pass

    def exit(self):
        pass

    def with_task(self, task: Task) -> 'Context':
        task.set_session(self)
        return self

    def fork(self, task: Task) -> 'Context':
        """fork a child message, typically when call a new tools"""
        return self.with_task(task)

    async def refresh_screenshot(self) -> str | None:
        """
        Refresh and return the latest screenshot of the running tools.

        Returns:
            None if no screenshot is available or the screenshot stays unchanged.
            Otherwise, return the latest screenshot.
        """
        if not self.__active_tool or self.is_finished():
            return None

        screenshot = await self.__active_tool.get_screenshot()

        if screenshot == self.__last_screenshot:
            return None

        self.__last_screenshot = screenshot

        return screenshot

    async def send(self, cb) -> None:
        """send a message to the context"""
        # self.cb_dict[cb.id()] = cb
        await self.q.put(cb)

    async def fetch(self):
        """receive a message"""
        while not self.is_failed() and not self.is_finished():
            try:
                item = self.q.get_nowait()
                if item:
                    return item
            except asyncio.QueueEmpty:
                await asyncio.sleep(0.01)

    def get_callback(self, cb_id: str):
        pass
        # return self.cb_dict[cb_id]

    def finish(self, msg: str):
        self.__result = msg
        self.__is_finished = True
        self.__last_screenshot = None
        self.q.task_done()

    def failed(self, msg: str):
        self.__failed_msg = msg
        self.__is_failed = True
        self.q.task_done()

    def is_finished(self) -> bool:
        return self.__is_finished

    def get_result(self) -> str:
        return self.__result

    def is_failed(self) -> bool:
        return self.__is_failed

    def get_failed_msg(self) -> str:
        return self.__failed_msg

    def retrieve(self, msg: str) -> str:
        pass

    def bind(self, tool):
        self.__active_tool = tool

    def get_tool(self):
        return self.__active_tool



from npiai.core.browser import PlaywrightContext, NavigatorAgent
from .hitl import HITL
from .base import BaseTool

__all__ = [
    'BaseTool',
    'PlaywrightContext',
    'NavigatorAgent',
    'HITL',
]



import asyncio

from npiai import agent
from npiai.tools import Slack
from npiai.hitl_handler import ConsoleHandler


async def main():
    async with agent.wrap(Slack()) as slack:
        slack.use_hitl(ConsoleHandler())

        return await slack.chat(
            'Send a direct message to Dolphin asking if he is doing well, and wait for his reply.'
        )


if __name__ == '__main__':
    asyncio.run(main())



import os
from typing import List, overload

from pydantic import create_model, Field

from npiai.llm import LLM, OpenAI
from npiai.types import FunctionRegistration
from npiai.context import Context, Task
from npiai.core.base import BaseAgentTool
from npiai.core.hitl import HITL
from npiai.core.tool._function import FunctionTool
from npiai.core.tool._browser import BrowserTool
from npiai.utils import sanitize_schema

from litellm.types.completion import (
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)


class AgentTool(BaseAgentTool):

    def __init__(self, tool: FunctionTool, llm: LLM = None):
        self.name = f'{tool.name}_agent'
        self._tool = tool
        self.llm = llm or OpenAI(api_key=os.environ.get('OPENAI_API_KEY', None), model='gpt-4o')
        self.description = tool.description
        self.provider = tool.provider
        super().__init__(tool.name, tool.description, tool.provider)

    def unpack_functions(self) -> List[FunctionRegistration]:
        # Wrap the chat function of this agent to FunctionRegistration
        model = create_model(
            f'{self.name}_agent_model',
            instruction=(str, Field(
                description=f'The task you want {self._tool.name} to do or '
                            f'the message you want to chat with {self._tool.name}'
            ))
        )

        fn_reg = FunctionRegistration(
            fn=self.chat,
            name='chat',
            ctx_param_name='ctx',
            description=f"This is an api of an AI Assistant, named {self.name}, the abilities of the assistant is:\n "
                        f"{self.description}\n"
                        f"You can use this function to direct the assistant to accomplish task for you.",
            model=model,
            schema=sanitize_schema(model),
        )

        return [fn_reg]

    def use_hitl(self, hitl: HITL):
        # super().use_hitl(hitl)
        self._tool.use_hitl(hitl)

    async def start(self):
        await self._tool.start()

    async def end(self):
        await self._tool.end()

    async def chat(
            self,
            ctx: Context,
            instruction: str,
    ) -> str:
        task = Task(goal=instruction)
        ctx = ctx.fork(task)
        if self._tool.system_prompt:
            await task.step([
                ChatCompletionSystemMessageParam(role='system', content=self._tool.system_prompt)
            ])

        await task.step([
            ChatCompletionUserMessageParam(role='user', content=instruction)
        ])
        return await self._call_llm(ctx, task)

    async def _call_llm(self, session: Context, task: Task) -> str:
        while True:
            response = await self.llm.completion(
                messages=task.conversations(),
                tools=self._tool.tools,
                tool_choice='auto',
                max_tokens=4096,
            )
            await task.step([response.choices[0].message])

            response_message = response.choices[0].message
            tool_calls = response_message.get('tool_calls', None)

            if tool_calls is None:
                return response_message.content

            results = await self._tool.call(tool_calls, session)
            await task.step(results)


class BrowserAgentTool(AgentTool):

    def __init__(self, tool: BrowserTool, llm: LLM = None):
        self._tool = tool
        super().__init__(tool, llm)

    async def get_screenshot(self) -> str | None:
        return await self._tool.get_screenshot()

    async def goto_blank(self):
        await self._tool.goto_blank()

    async def chat(
            self,
            ctx: Context,
            instruction: str,
    ) -> str:
        if not self._tool.use_screenshot:
            return await super().chat(ctx, instruction)

        screenshot = await self._tool.get_screenshot()

        if not screenshot:
            return await super().chat(ctx, instruction)

        task = Task(goal=instruction)
        ctx = ctx.fork(task)
        if self._tool.system_prompt:
            await task.step([ChatCompletionSystemMessageParam(role='system', content=self._tool.system_prompt)])

        await task.step([ChatCompletionUserMessageParam(
            role='user',
            content=[
                {
                    'type': 'text',
                    'text': instruction,
                },
                {
                    'type': 'image_url',
                    'image_url': {
                        'url': screenshot,
                    },
                },
            ])])

        return await self._call_llm(ctx, task)


@overload
def agent_wrapper(tool: FunctionTool, llm: LLM = None) -> AgentTool:
    ...


@overload
def agent_wrapper(tool: BrowserTool, llm: LLM = None) -> BrowserAgentTool:
    ...


def agent_wrapper(tool: FunctionTool | BrowserTool, llm: LLM = None) -> AgentTool | BrowserAgentTool:
    if isinstance(tool, BrowserTool):
        return BrowserAgentTool(tool, llm)

    if isinstance(tool, FunctionTool):
        return AgentTool(tool, llm)

    raise TypeError(f'app must be an instance of FunctionTool or BrowserTool')



from .context import Context

from playground.proto import playground_pb2


class ContextManager:
    """the manager of the context"""

    def __init__(self):
        self.threads = {}

    def new_thread(self, req: playground_pb2.ChatRequest) -> Context:
        """create a context"""
        th = Context()
        self.threads[th.id] = th
        return th

    def get_thread(self, tid: str) -> Context:
        """get a context by id"""
        return self.threads.get(tid)

    def release(self, tid: str) -> None:
        """delete a context by id"""
        del self.threads[tid]



"""The basic interface for NPi Apps"""
import dataclasses
import inspect
import json
import re
from typing import Dict, List, Optional, Any, Type

import yaml

from pydantic import Field, create_model
from litellm.types.completion import ChatCompletionToolMessageParam, ChatCompletionMessageParam
from litellm.types.utils import ChatCompletionMessageToolCall
from openai.types.chat import ChatCompletionToolParam

from npiai.core.base import BaseTool, BaseFunctionTool
from npiai.core.hitl import HITL
from npiai.types import FunctionRegistration, ToolFunction, Shot, ToolMeta
from npiai.utils import logger, sanitize_schema, parse_docstring, to_async_fn
from npiai.context import Context
from npiai.core import callback

__NPI_TOOL_ATTR__ = '__NPI_TOOL_ATTR__'


def function(
        tool_fn: ToolFunction = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        schema: Dict[str, Any] = None,
        model: Optional[Type[BaseTool]] = None,
        few_shots: Optional[List[Shot]] = None,
):
    """
    NPi Tool decorator for functions

    Args:
        tool_fn: Tool function. This value will be set automatically.
        name: Tool name. The tool function name will be used if not given.
        description: Tool description. This value will be inferred from the tool's docstring if not given.
        schema: Tool parameters schema. This value will be inferred from the tool's type hints if not given.
        model: Pydantic model for tool schema. This value will be built from the schema if not given.
        few_shots: Predefined working examples.

    Returns:
        Wrapped tool function that will be registered on the tools class
    """

    def decorator(fn: ToolFunction):
        setattr(
            fn,
            __NPI_TOOL_ATTR__,
            ToolMeta(
                name=name,
                description=description,
                schema=schema,
                model=model,
                few_shots=few_shots,
            )
        )

        return fn

    # called as `@function`
    if callable(tool_fn):
        return decorator(tool_fn)

    # called as `@function(...)`
    return decorator


class FunctionTool(BaseFunctionTool):
    """The basic interface for the natural language programming interface"""

    name: str
    description: str
    system_prompt: Optional[str]
    provider: str

    _tools: List[ChatCompletionToolParam]
    _fn_map: Dict[str, FunctionRegistration]
    _sub_tools: List[BaseTool]

    _started: bool = False

    @property
    def functions(self):
        return list(self._fn_map.values())

    @property
    def sub_tools(self):
        return self._sub_tools

    def __init__(
            self,
            name: str,
            description: str,
            system_prompt: str = None,
            provider: str = None,
    ):
        self.name = name
        self.description = description
        self.system_prompt = system_prompt
        self.provider = provider or 'private'
        self._tools = []
        self._fn_map = {}
        self._sub_tools = []

        self._register_tools()
        super().__init__(self.name, self.description, self.provider, self._fn_map, )

    def unpack_functions(self) -> List[FunctionRegistration]:
        return list(self._fn_map.values())

    def use_hitl(self, hitl: HITL):
        """
        Attach the given HITL handler to this tool and all its sub apps

        Args:
            hitl: HITL handler
        """
        super().use_hitl(hitl)

        for app in self._sub_tools:
            app.use_hitl(hitl)

    async def start(self):
        """Start the tools"""
        if not self._started:
            self._started = True
            for tool in self._sub_tools:
                await tool.start()

    async def end(self):
        """Stop and dispose the tools"""
        if self._started:
            self._started = False
            for app in self._sub_tools:
                await app.end()

    def add_tool(
            self,
            *tools: BaseTool,
    ):
        for tool in tools:
            # share hitl handler
            if tool._hitl is None:
                tool.use_hitl(self._hitl)

            self._sub_tools.append(tool)

            for fn_reg in tool.unpack_functions():
                scoped_fn_reg = dataclasses.replace(fn_reg, name=f'{tool.name}_{fn_reg.name}')
                self._add_function(scoped_fn_reg)

    async def debug(self, session: Context | None = None, app_name: str = None, fn_name: str = None,
                    args: Dict[str, Any] = None) -> str:
        if app_name:
            fn_name = f'{app_name}_{fn_name}'

        return await self.exec(session, fn_name, args)

    async def call(
            self,
            tool_calls: List[ChatCompletionMessageToolCall],
            session: Context | None = None,
    ) -> List[ChatCompletionMessageParam]:
        if session is None:
            session = Context()
        results: List[ChatCompletionToolMessageParam] = []

        for call in tool_calls:
            fn_name = call.function.name
            args = json.loads(call.function.arguments)
            call_msg = f'[{self.name}]: Calling {fn_name}'

            if args:
                arg_list = ', '.join(f'{k}={json.dumps(v)}' for k, v in args.items())
                call_msg += f'({arg_list})'
            else:
                call_msg += '()'

            logger.info(call_msg)
            await session.send(call_msg)

            try:
                res = await self.exec(session, fn_name, args)
            except Exception as e:
                logger.error(e)
                res = f'Exception while executing {fn_name}: {e}'
                await session.send(res)

            logger.debug(f'[{self.name}]: function `{fn_name}` returned:\n {res}')

            results.append(
                ChatCompletionToolMessageParam(
                    role='tool',
                    name=fn_name,
                    tool_call_id=call.id,
                    content=str(res),
                )
            )

        return results

    def _register_tools(self):
        """
        Find the wrapped tool functions and register them in this tools
        """
        for attr, fn in inspect.getmembers(self, lambda x: inspect.ismethod(x) or inspect.isfunction(x)):
            tool_meta: ToolMeta | None = getattr(fn, __NPI_TOOL_ATTR__, None)

            if not callable(fn) or not tool_meta:
                continue

            params = list(inspect.signature(fn).parameters.values())
            docstr = parse_docstring(inspect.getdoc(fn))
            tool_name = tool_meta.name or fn.__name__
            tool_desc = tool_meta.description or docstr.description

            if not tool_desc:
                raise ValueError(
                    f'Unable to get the description of tool function `{fn}`'
                )

            # parse schema
            tool_model = tool_meta.model
            tool_schema = tool_meta.schema
            ctx_param_name = None

            if not tool_model and len(params) > 0:
                # get parameter descriptions
                param_descriptions = {}
                for p in docstr.params:
                    param_descriptions[p.arg_name] = p.description

                # get parameter field definitions
                param_fields = {}
                for p in params:
                    if p.annotation is Context:
                        ctx_param_name = p.name
                        continue
                    param_fields[p.name] = (p.annotation, Field(
                        default=p.default if p.default is not inspect.Parameter.empty else ...,
                        description=param_descriptions.get(p.name, ''),
                    ))

                tool_model = create_model(f'{tool_name}_model', **param_fields)

            if not tool_schema and tool_model:
                tool_schema = sanitize_schema(tool_model)

            # parse examples
            tool_few_shots = tool_meta.few_shots
            doc_shots = [m for m in docstr.meta if m.args == ['few_shots']]

            if not tool_few_shots and len(doc_shots) > 0:
                tool_few_shots = []

                for shot in doc_shots:
                    items = re.findall(r'^\s*- ', shot.description, re.MULTILINE)
                    if len(items) == 1:
                        # remove leading '- ' to avoid indentation issues
                        shot_data = yaml.safe_load(re.sub(r'^\s*- ', '', shot.description))
                    else:
                        shot_data = yaml.safe_load(shot.description)

                    if not isinstance(shot_data, list):
                        shot_data = [shot_data]

                    for d in shot_data:
                        tool_few_shots.append(Shot(**d))

            self._add_function(
                FunctionRegistration(
                    # wrap fn in an async wrapper
                    fn=to_async_fn(fn),
                    name=tool_name,
                    ctx_param_name=ctx_param_name,
                    description=tool_desc.strip(),
                    schema=tool_schema,
                    model=tool_model,
                    few_shots=tool_few_shots,
                )
            )

    def _add_function(self, fn_reg: FunctionRegistration):
        if fn_reg.name in self._fn_map:
            raise Exception(f'Duplicate function: {fn_reg.name}')

        self._fn_map[fn_reg.name] = fn_reg
        self._tools.append(fn_reg.get_tool_param())



import asyncio

from npiai import agent
from npiai.tools import GitHub


async def main():
    async with agent.wrap(GitHub()) as github:
        return await github.chat('Create a test issue in idiotWu/npi-test with label "NPi" and assign it to @idiotWu')


if __name__ == '__main__':
    asyncio.run(main())
